;; Generated package description from compat.el  -*- no-byte-compile: t; lexical-binding:t -*-
(define-package "compat" "31.0.0.2.0.20260812.3" "Emacs Lisp Compatibility Library" '((emacs "25.1")) :commit "ad2720457067b72859ba1247f6fdb3361b204bc6" :authors '(("Philip Kaludercic" . "philipk@posteo.net") ("Daniel Mendler" . "mail@daniel-mendler.de")) :maintainer '(("Philip Kaludercic" . "philipk@posteo.net") ("Daniel Mendler" . "mail@daniel-mendler.de")) :keywords '("lisp" "maint") :url "https://github.com/emacs-compat/compat")
